# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2017-2021 TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
import aiohttp
import asyncio
import logging
from typing import Union, List

from .cooldowns import RateBucket
from .errors import HTTPException, Unauthorized


log = logging.getLogger(__name__)


class HTTPSession:
    BASE = 'https://api.twitch.tv/helix'
    TOKEN_BASE = "https://id.twitch.tv/oauth2/token"

    def __init__(self, loop, **attrs):
        self.client_id = client_id = attrs.get('client_id', None)
        self.client_secret = attrs.get("client_secret", None)
        self.token = attrs.get("api_token", None)
        self.scopes = attrs.get("scopes", [])

        if not client_id:
            log.warning('Running without client ID, HTTP endpoints will not work without authentication.')

        if not self.token and not client_id:
            log.warning("Running without client ID or bearer token, HTTP endpoints will not work without authentication.")

        self._bucket = RateBucket(method='http')
        self._session = aiohttp.ClientSession(loop=loop)
        self._refresh_token = None

    async def generate_token(self):
        if not self.client_id or not self.client_secret:
            raise HTTPException("Unable to generate a token, client id and/or client secret not given")

        if self._refresh_token:
            url = self.TOKEN_BASE + "?grant_type=refresh_token&refresh_token={0}&client_id={1}&client_secret={2}".format(
                self._refresh_token, self.client_id, self.client_secret)

        else:
            url = self.TOKEN_BASE + "?client_id={0}&client_secret={1}&grant_type=client_credentials".format(self.client_id, self.client_secret)
            if self.scopes:
                url += "&scope=" + " ".join(self.scopes)

        async with self._session.post(url) as resp:
            if 300 < resp.status or resp.status < 200:
                raise HTTPException("Unable to generate a token: " + await resp.text())

            data = await resp.json()
            self.token = data['access_token']
            self._refresh_token = data.get('refresh_token', None)
            logging.info("Invalid or no token found, generated new token: %s", self.token)

    async def request(self, method, url, *, params=None, limit=None, **kwargs):
        count = kwargs.pop('count', False)

        data = []

        params = params or []
        url = f'{self.BASE}{url}'

        # headers = {}

        headers = kwargs.pop('headers', {})

        if self.client_id is not None:
            headers['Client-ID'] = str(self.client_id)

        if self.client_secret and self.client_id and not self.token:
            logging.info("No token passed, generating new token under client id {0} and client secret {1}".format(self.client_id, self.client_secret))
            await self.generate_token()

        if self.token is not None and "Authorization" not in headers:
            headers['Authorization'] = "Bearer " + self.token

        #else: we'll probably get a 401, but we can check this in the response

        cursor = None

        def reached_limit():
            return limit and len(data) >= limit

        def get_limit():
            if limit is None:
                return '100'

            to_get = limit - len(data)
            return str(to_get) if to_get < 100 else '100'

        is_finished = False
        while not is_finished:
            if limit is not None:
                if cursor is not None:
                    params.append(('after', cursor))

                params.append(('first', get_limit()))

            body, is_text = await self._request(method, url, params=params, headers=headers, **kwargs)
            if is_text:
                return body

            if count:
                return body['total']

            params.pop()  # remove the first param

            if cursor is not None:
                params.pop()

            data += body['data']

            try:
                cursor = body['pagination'].get('cursor', None)
            except KeyError:
                break
            else:
                if not cursor:
                    break

            is_finished = reached_limit() if limit is not None else True

        return data

    async def _request(self, method, url, utilize_bucket=True, **kwargs):
        reason = None

        for attempt in range(5):
            if utilize_bucket and self._bucket.limited:
                await self._bucket.wait_reset()

            async with self._session.request(method, url, **kwargs) as resp:
                if 500 <= resp.status <= 504:
                    reason = resp.reason
                    await asyncio.sleep(2 ** attempt + 1)
                    continue

                if utilize_bucket:
                    reset = resp.headers.get('Ratelimit-Reset')
                    remaining = resp.headers.get('Ratelimit-Remaining')

                    self._bucket.update(reset=reset, remaining=remaining)

                if 200 <= resp.status < 300:
                    if resp.content_type == 'application/json':
                        return await resp.json(), False

                    return await resp.text(encoding='utf-8'), True

                if resp.status == 401:
                    if self.client_id is None:
                        raise Unauthorized('A client ID and Bearer token is needed to use this route.')

                    if "WWW-Authenticate" in resp.headers:
                        try:
                            await self.generate_token()
                        except:
                            raise Unauthorized("Your oauth token is invalid, and a new one could not be generated")

                    raise Unauthorized('You\'re not authorized to use this route.')

                if resp.status == 429:
                    reason = 'Ratelimit Reached'

                    if not utilize_bucket:  # non Helix APIs don't have ratelimit headers
                        await asyncio.sleep(3 ** attempt + 1)
                    continue

                raise HTTPException(f'Failed to fulfil request ({resp.status}).', resp.reason, resp.status)

        raise HTTPException('Failed to reach Twitch API', reason, resp.status)

    @staticmethod
    def _populate_entries(*channels: Union[str, int]):
        names = set()
        ids = set()

        for channel in channels:
            if isinstance(channel, str):
                if channel.isdigit():
                    # Handle ids in the string form
                    ids.add(int(channel))
                else:
                    names.add(channel)
            elif isinstance(channel, int):
                ids.add(str(channel))

        if len(names | ids) > 100:
            raise HTTPException('Bad Request - Total entries must not exceed 100.')

        return names, ids

    async def get_users(self, *users: Union[str, int]):
        names, ids = self._populate_entries(*users)
        params = [('id', x) for x in ids] + [('login', x) for x in names]

        return await self.request('GET', '/users', params=params)

    async def get_follow(self, from_id: str, to_id: str):
        params = [('from_id', from_id), ('to_id', to_id)]
        return await self.request('GET', '/users/follows', params=params)

    async def get_followers(self, user_id: str, *, count):
        params = [('to_id', user_id)]
        return await self.request('GET', '/users/follows', params=params, count=count)

    async def get_following(self, user_id: str, *, count):
        params = [('from_id', user_id)]
        return await self.request('GET', '/users/follows', params=params, count=count)

    async def get_streams(self, *, game_id=None, language=None, channels, limit=None):
        if channels:
            names, ids = self._populate_entries(*channels)
            params = [('user_id', x) for x in ids] + [('user_login', x) for x in names]
        else:
            params = []

        if game_id is not None:
            params.append(('game_id', str(game_id)))

        if language is not None:
            params.append(('language', language))

        return await self.request('GET', '/streams', params=params, limit=limit)

    async def get_games(self, *games: Union[str, int]):
        names, ids = self._populate_entries(*games)
        params = [('id', x) for x in ids] + [('name', x) for x in names]

        return await self.request('GET', '/games', params=params)

    async def get_top_games(self, limit=None):
        return await self.request('GET', '/games/top', limit=limit)

    async def modify_webhook_subscription(self, *, callback, mode, topic, lease_seconds, secret=None):
        data = {
            'hub.callback': callback,
            'hub.mode': mode,
            'hub.topic': topic,
            'hub.lease_seconds': lease_seconds
        }

        if secret is not None:
            data['secret'] = secret

        return await self.request('POST', '/webhooks/hub', json=data)

    async def create_clip(self, token: str, broadcaster_id: int):
        params = [('broadcaster_id', str(broadcaster_id))]
        return await self.request('POST', '/clips', params=params, headers={'Authorization': f'Bearer {token}'})

    async def create_reward(self,
                            token: str,
                            broadcaster_id: int,
                            title: str,
                            cost: int,
                            prompt: str = None,
                            is_enabled: bool = True,
                            background_color: str = None,
                            user_input_required: bool = False,
                            max_per_stream: int = None,
                            max_per_user: int = None,
                            global_cooldown: int = None,
                            fufill_immediatly: bool = False
                            ):
        params = [("broadcaster_id", str(broadcaster_id))]
        data = {
            "title": title,
            "cost": cost,
            "prompt": prompt,
            "is_enabled": is_enabled,
            "is_user_input_required": user_input_required,
            "should_redemptions_skip_request_queue": fufill_immediatly
        }
        if max_per_stream:
            data['max_per_stream'] = max_per_stream
            data['max_per_stream_enabled'] = True

        if max_per_user:
            data['max_per_user_per_stream'] = max_per_user
            data['max_per_user_per_stream_enabled'] = True

        if background_color:
            data['background_color'] = background_color

        if global_cooldown:
            data['global_cooldown_seconds'] = global_cooldown
            data['is_global_cooldown_enabled'] = True

        return await self.request('POST', '/channel_points/custom_rewards', params=params, json=data, headers={"Authorization": f"Bearer {token}"})

    async def get_rewards(self, token: str, broadcaster_id: int, only_manageable: bool = False, ids: List[int]=None):
        params = [("broadcaster_id", str(broadcaster_id)), ("only_manageable_rewards", str(only_manageable))]

        if ids:
            for id in ids:
                params.append(("id", str(id)))

        return await self.request("GET", "/channel_points/custom_rewards", params=params, headers={"Authorization": f"Bearer {token}"})

    async def update_reward(
                            self,
                            token: str,
                            broadcaster_id: int,
                            reward_id: str,
                            title: str = None,
                            prompt: str = None,
                            cost: int = None,
                            background_color: str = None,
                            enabled: bool = None,
                            input_required: bool = None,
                            max_per_stream_enabled: bool = None,
                            max_per_stream: int = None,
                            max_per_user_per_stream_enabled: bool = None,
                            max_per_user_per_stream: int = None,
                            global_cooldown_enabled: bool = None,
                            global_cooldown: int = None,
                            paused: bool = None,
                            redemptions_skip_queue: bool = None
                            ):
        data = {}
        if title:
            data['title'] = title

        if prompt:
            data['prompt'] = prompt

        if cost:
            data['cost'] = cost

        if background_color:
            data['background_color'] = background_color

        if enabled is not None:
            data['enabled'] = enabled

        if input_required is not None:
            data['is_user_input_required'] = input_required

        if max_per_stream_enabled is not None:
            data['is_max_per_stream_enabled'] = max_per_stream_enabled

        if max_per_stream is not None:
            data['max_per_stream'] = max_per_stream

        if max_per_user_per_stream_enabled is not None:
            data['is_max_per_user_per_stream_enabled'] = max_per_user_per_stream_enabled

        if max_per_user_per_stream is not None:
            data['max_per_user_per_stream'] = max_per_user_per_stream

        if global_cooldown_enabled is not None:
            data['is_global_cooldown_enabled'] = global_cooldown_enabled

        if global_cooldown is not None:
            data['global_cooldown_seconds'] = global_cooldown

        if paused is not None:
            data['is_paused'] = paused

        if redemptions_skip_queue is not None:
            data['should_redemptions_skip_request_queue'] = redemptions_skip_queue

        if not data:
            raise ValueError("Nothing changed!")

        params = [("broadcaster_id", str(broadcaster_id)), ("id", str(reward_id))]
        return await self.request("PATCH", "/channel_points/custom_rewards", params=params, headers={"Authorization": f"Bearer {token}"}, json=data)

    async def delete_custom_reward(self, token: str, broadcaster_id: int, reward_id: str):
        params = [("broadcaster_id", str(broadcaster_id), ("id", reward_id))]
        await self.request("DELETE", "channel_points/custom_rewards", params=params, headers={"Authorization": f"Bearer {token}"})

    async def get_reward_redemptions(self, token: str, broadcaster_id: int, reward_id: str,
                                           redemption_id: str = None, status: str = None, sort: str = None, limit: int = 100):
        params = [("broadcaster_id", str(broadcaster_id)), ("reward_id", reward_id)]

        if redemption_id:
            params.append(("id", redemption_id))

        if status:
            params.append(("status", status))

        if sort:
            params.append(("sort", sort))

        return await self.request("GET", "/channel_points/custom_rewards/redemptions", params=params, headers={"Authorization": f"Bearer {token}"}, limit=limit)

    async def update_reward_redemption_status(self, token: str, broadcaster_id: int, reward_id: str, custom_reward_id: str, status: bool):
        params = [("id", custom_reward_id), ("broadcaster_id", str(broadcaster_id)), ("reward_id", reward_id)]
        status = "FULFILLED" if status else "CANCELLED"
        return await self.request("PATCH", "/channel_points/custom_rewards/redemptions", params=params,
                                  json={"status": status}, headers={"Authorization": f"Bearer {token}"})



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2017-2021 TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import abc
import time

from .cooldowns import RateBucket
from .errors import *


class IRCLimiterMapping:

    def __init__(self):
        self.buckets = {}

    def get_bucket(self, channel: str, method: str) -> RateBucket:
        try:
            bucket = self.buckets[channel]
        except KeyError:
            bucket = RateBucket(method=method)
            self.buckets[channel] = bucket

        if bucket.method != method:
            bucket.method = method
            if method == 'mod':
                bucket.limit = bucket.MODLIMIT
            else:
                bucket.limit = bucket.IRCLIMIT

            self.buckets[channel] = bucket

        return bucket


limiter = IRCLimiterMapping()


class Messageable(metaclass=abc.ABCMeta):

    __slots__ = ()

    __invalid__ = ('ban', 'unban', 'timeout', 'w', 'colour', 'color', 'mod',
                   'unmod', 'clear', 'subscribers', 'subscriberoff', 'slow', 'slowoff',
                   'r9k', 'r9koff', 'emoteonly', 'emoteonlyoff', 'host', 'unhost')

    @abc.abstractmethod
    def _get_channel(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def _get_socket(self):
        raise NotImplementedError

    @abc.abstractmethod
    def _get_method(self):
        raise NotImplementedError

    def check_bucket(self, channel):
        ws = self._get_socket

        try:
            bot = ws._channel_cache[channel]['bot']
        except KeyError:
            bucket = limiter.get_bucket(channel=channel, method='irc')
        else:
            if bot.is_mod:
                bucket = limiter.get_bucket(channel=channel, method='mod')
            else:
                bucket = limiter.get_bucket(channel=channel, method='irc')

        now = time.time()
        bucket.update()

        if bucket.limited:
            raise TwitchIOBException(f'IRC Message rate limit reached for channel <{channel}>.'
                                     f' Please try again in {bucket._reset - now:.2f}s')

    @staticmethod
    def check_content(channel, content: str):
        if not channel:
            raise TwitchIOBException('Invalid channel for Messageable. Must be channel or user.')

        if len(content) > 500:
            raise InvalidContent('Length of message can not be > 500.')

    async def send(self, content: str):
        """|coro|

        Send a message to the destination associated with the dataclass.

        Destination will either be a channel or user.
        Chat commands are not allowed to be invoked with this method.

        Parameters
        ------------
        content: str
            The content you wish to send as a message. The content must be a string.

        Raises
        --------
        TwitchIOBException
            Invalid destination.
        InvalidContent
            Invalid content.
        """
        content = str(content)

        channel, user = self._get_channel()
        method = self._get_method()

        self.check_content(channel, content)

        if content.startswith(('.', '/')):
            if content.lstrip('./').startswith(self.__invalid__):
                raise InvalidContent('UnAuthorised chat command for send. Use built in method(s).')

        ws = self._get_socket
        self.check_bucket(channel)

        if method == 'User':
            content = f'.w {user} {content}'

        await ws.send_privmsg(channel, content=content)

    async def clear(self):
        """|coro|

        Method which sends .clear to Twitch and clears the chat.
        """

        ws = self._get_socket
        channel, _ = self._get_channel()

        self.check_bucket(channel)

        await ws.send_privmsg(channel, content=f'.clear')

    async def slow(self):
        """|coro|

        Method which sends a .slow to Twitch and sets the channel to slowmode.
        """

        ws = self._get_socket
        channel, _ = self._get_channel()

        self.check_bucket(channel)

        await ws.send_privmsg(channel, content=f'.slow')

    async def unslow(self):
        """|coro|

        Method which sends a .slowoff to Twitch and sets the channel to slowmode off.
        """

        ws = self._get_socket
        channel, _ = self._get_channel()

        self.check_bucket(channel)

        await ws.send_privmsg(channel, content=f'.slowoff')

    async def slow_off(self):
        """|coro|

        Alias to unslow.
        """
        await self.unslow()

    async def timeout(self, user: str, duration: int=600, reason: str=''):
        """|coro|

        Method which sends a .timeout command to Twitch.

        Parameters
        ------------
        user: str
            The user you wish to timeout.
        duration: int
            The duration in seconds to timeout the user.
        reason: Optional[str]
            The reason you timed out the user.
        """

        ws = self._get_socket
        channel, _ = self._get_channel()

        self.check_bucket(channel)

        await ws.send_privmsg(channel, content=f'.timeout {user} {duration} {reason}')

    async def ban(self, user: str, reason: str=''):
        """|coro|

        Method which sends a .ban command to Twitch.

        Parameters
        ------------
        user: str
            The user you would like to ban.
        reason: Optional[str]
            The reason you banned this user.
        """

        ws = self._get_socket
        channel, _ = self._get_channel()

        self.check_bucket(channel)

        await ws.send_privmsg(channel, content=f'.ban {user} {reason}')

    async def unban(self, user: str):
        """|coro|

        Method which sends a .unban command to Twitch.

        Parameters
        ------------
        user: str
            The user you wish to unban.
        """

        ws = self._get_socket
        channel, _ = self._get_channel()

        self.check_bucket(channel)

        await ws.send_privmsg(channel, content=f'.unban {user}')

    async def send_me(self, content: str):
        """|coro|

        Method which sends .me along with your content.

        Parameters
        ------------
        content: str
            The message you wish to send.

        Raises
        --------
        InvalidContent
            The content exceeded 500 characters.
        """

        ws = self._get_socket
        channel, _ = self._get_channel()

        self.check_bucket(channel)
        self.check_content(channel, content)

        await ws.send_privmsg(channel, content=f'.me {content}')

    async def colour(self, colour: str):
        """Send a colour change request to Twitch.

        Parameters
        ------------
        colour: str
            A predefined colour listed below. Turbo Users may use a valid hex code. e.g #233233

            Blue
            BlueViolet
            CadetBlue
            Chocolate
            Coral
            DodgerBlue
            FireBrick
            GoldenRod
            Green
            HotPink
            OrangeRed
            Red
            SeaGreen
            SpringGreen
            YellowGreen
        """
        ws = self._get_socket
        channel, _ = self._get_channel()

        self.check_bucket(channel)
        await ws.send_privmsg(channel, content=f'.color {colour}')

    async def color(self, colour: str):
        """An alias to colour."""
        await self.colour(colour)




# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2017-2021 TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import asyncio
import importlib
import inspect
import itertools
import sys
import threading
import traceback
import uuid
from typing import Union, List, Tuple

from .core import Command, AutoCog
from .errors import *
from .stringparser import StringParser
from twitchio.client import Client
from twitchio.dataclasses import Context
from twitchio.errors import ClientError
from twitchio.webhook import TwitchWebhookServer
from twitchio.websocket import WebsocketConnection


class Bot(Client):
    """Twitch IRC Bot.

    Parameters
    ------------
    irc_token: str
        The OAuth token to use for IRC.
    client_id: str:
        Your application ID, used for HTTP endpoints.
    prefix: str
        The bots prefix.
    nick: str
        The bots nick in lowercase.
    loop: [Optional]
        The asyncio event loop to use.
    initial_channels: list
        The initial channels for the bot to join on startup.
    webhook_server: bool [Optional]
        A bool indicating whether the built-in webhook server should be used.
    local_host: str [Optional]
        The local host the webhook server should run on.
    external_host: str [Optional]
        The external address the webhook_server should lsiten on.
    port: int [Optional]
        The port the webhook_server should be started on.
    callback: str [Optional]
        The pages as a string where the webhook_server should lsiten for events.

    Notes
    -------
    .. note::

        To enable the webhook server, the webhook_server parameter must be True.
        A local_host, external_host and port must also be provided.


        An optional parameter `callback` may be passed. This should be the page Twitch sends data to.
        A long random string, such as hex, is advised e.g `2t389hth892t3h898hweiogtieo`
    """

    def __init__(self, irc_token: str, api_token: str=None, *, client_id: str=None, client_secret: str=None, prefix: Union[list, tuple, str],
                 nick: str, loop: asyncio.BaseEventLoop=None, initial_channels: Union[list, tuple]=None,
                 webhook_server: bool=False, local_host: str=None, external_host: str=None, callback: str=None,
                 port: int=None, **attrs):

        self.loop = loop or asyncio.get_event_loop()
        super().__init__(loop=self.loop, client_id=client_id, api_token=api_token, client_secret=client_secret, **attrs)
        self.nick = nick
        self.initial_channels = initial_channels

        self._ws = WebsocketConnection(bot=self, loop=self.loop, http=self.http, irc_token=irc_token,
                                       nick=nick, initial_channels=initial_channels, **attrs)

        self._webhook_server = None
        if webhook_server:
            self._webhook_server = TwitchWebhookServer(bot=self,
                                                       local=local_host,
                                                       external=external_host,
                                                       callback=callback,
                                                       port=port)
            loop = asyncio.new_event_loop()
            thread = threading.Thread(target=self._webhook_server.run_server, args=(loop, ), daemon=True)
            thread.start()

        self.loop.create_task(self._prefix_setter(prefix))

        self.extra_listeners = {}
        self.commands = {}
        self.modules = {}
        self.cogs = {}
        self._aliases = {}
        self._checks = []
        self.prefixes = None

        self._init_methods()

    def _init_methods(self):
        commands = inspect.getmembers(self)

        for _, obj in commands:
            if not isinstance(obj, Command):
                continue

            obj.instance = self

            try:
                self.add_command(obj)
            except CommandError:
                traceback.print_exc()
                continue

    def add_command(self, command):
        if not isinstance(command, Command):
            raise TypeError('Commands passed my be a subclass of Command.')
        elif command.name in self.commands:
            raise CommandError(f'Failed to load command <{command.name}>, a command with that name already exists')
        elif not inspect.iscoroutinefunction(command._callback):
            raise CommandError(f'Failed to load command <{command.name}>. Commands must be coroutines.')

        self.commands[command.name] = command

        if not command.aliases:
            return

        for alias in command.aliases:
            if alias in self.commands:
                del self.commands[command.name]
                raise CommandError(
                    f'Failed to load command <{command.name}>, a command with that name/alias already exists.')

            self._aliases[alias] = command.name

    def remove_command(self, command):
        if command.aliases:
            for a in command.aliases:
                self._aliases.pop(a)

        try:
            del self.commands[command.name]
        except KeyError:
            # Not sure why this would happen, but people be people.
            pass

    def load_module(self, name: str):
        """Method which loads a module and it's cogs.

        Parameters
        ------------
        name: str
            The name of the module to load in dot.path format.
        """
        if name in self.modules:
            return

        valid = False

        module = importlib.import_module(name)
        for _, member in inspect.getmembers(module):
            if inspect.isclass(member) and issubclass(member, AutoCog):
                member(self)._prepare(self)
                valid = True

        if hasattr(module, 'prepare'):
            module.prepare(self)
        elif not valid:
            del module
            del sys.modules[name]
            raise ImportError(f'Module <{name}> is missing a prepare method')

        if name not in self.modules:
            self.modules[name] = module

    def unload_module(self, name: str):
        """Method which unloads a module and it's cogs/commands/events.

        Parameters
        ------------
        name: str
            The name of the module to load in dot.path format.
        """
        module = self.modules.pop(name, None)
        if not module:
            return

        cogs = []
        for cogname, c in self.cogs.copy().items():
            if c.__module__ == name:
                cogs.append(c.__name__)

        try:
            module.breakdown(self)
        finally:
            del module
            del sys.modules[name]

    def add_cog(self, cog):
        """Method which loads a cog and adds it's commands and events.

        Parameters
        ------------
        cog:
            An instance of the cog you wish to load.
        """
        members = inspect.getmembers(cog)

        for name, member in members:
            if isinstance(member, Command):
                member.instance = cog
                self.add_command(member)
            elif name.startswith('event_'):
                self.add_listener(member, name)

        self.cogs[type(cog).__name__] = cog

    def remove_cog(self, cogname: str):
        """Method which removes a cog and adds it's commands and events.

        Parameters
        ------------
        cogname:
            The name of the cog you wish to remove.
        """
        cog = self.cogs.pop(cogname, None)
        if not cog:
            return

        for name, member in inspect.getmembers(cog):
            if isinstance(member, Command):
                self.remove_command(member)
            elif name.startswith('event_'):
                del self.extra_listeners[name]
            elif name in self.extra_listeners:
                del self.extra_listeners[member.__name__]

        try:
            unload = getattr(cog, f'_{cog.__name__}__unload')
        except AttributeError:
            pass
        else:
            unload(self)

        del cog

    def add_check(self, func):
        """Adds a global check to the bot.

        Parameters
        ------------
        func : callable
            The function or coroutine to add as a global check to the bot.
        """

        self._checks.append(func)

    def remove_check(self, func):
        """Remove a global check from the bot.

        Parameters
        ------------
        func : callable
            The function to remove as a global check from the bot.
        """

        self._checks.remove(func)

    def run(self):
        """A blocking call that initializes the IRC Bot event loop.

        This should be the last function to be called.

        .. warning::
            You do not need to use this function unless you are accessing the IRC Endpoints.
        .. warning::
            You do not use this function if you are using :meth:`.start`
        """
        loop = self.loop or asyncio.get_event_loop()

        loop.run_until_complete(self._ws._connect())

        try:
            loop.run_until_complete(self._ws._listen())
        except KeyboardInterrupt:
            pass
        finally:
            self._ws.teardown()

    async def start(self):
        """|coro|

        An asynchronous call which starts the IRC Bot event loop.

        This should only be used when integrating Twitch Bots with Discord Bots.
        :meth:`.run` should be used instead.

        .. warning::
            Do not use this function if you are using :meth:`.run`
        """
        await self._ws._connect()

        try:
            await self._ws._listen()
        except KeyboardInterrupt:
            pass
        finally:
            self._ws.teardown()

    async def _prefix_setter(self, item):
        if inspect.iscoroutinefunction(item):
            item = await item()
        elif callable(item):
            item = item()

        if isinstance(item, (list, tuple)):
            self.prefixes = item
        elif isinstance(item, str):
            self.prefixes = [item]
        else:
            raise ClientError('Invalid prefix provided. A list, tuple, str or callable returning either should be used.')

    async def _get_prefixes(self, message):
        prefix = ret = self.prefixes
        if callable(prefix):
            ret = prefix(self, message.content)
            if inspect.isawaitable(ret):
                ret = await ret

        if isinstance(ret, (list, tuple)):
            ret = [p for p in ret if p]

        if isinstance(ret, str):
            ret = [ret]

        if not ret:
            raise ClientError('Invalid prefix provided.')

        return ret

    async def get_prefix(self, message):
        prefixes = await self._get_prefixes(message)

        prefix = None
        content = message.content

        for pre in prefixes:
            if content.startswith(pre):
                prefix = pre
                break

        return prefix

    def get_channel(self, name: str):
        """Retrieves a :class:`.Channel` from cache.

        Parameters
        ------------
        name: str
            The channel name to retrieve from cache.
        """
        cache = self._ws._channel_cache.get(name.lower())
        if cache:
            return cache['channel']

        return None

    async def join_channels(self, channels: Union[List[str], Tuple[str]]):
        """|coro|

        Join the specified channels.

        Parameters
        ------------
        channels: Union[List[str], Tuple[str]]
            The channels in either a list or tuple form to join.
        """
        await self._ws.join_channels(*channels)

    async def part_channels(self, channels: Union[List[str], Tuple[str]]):
        """|coro|

        Part the specified channels.

        Parameters
        ------------
        channels: Union[List[str], Tuple[str]]
            The channels in either a list or tuple form to part.
        """
        await self._ws.part_channels(*channels)

    async def get_context(self, message, cls=None):
        """|coro|

        A function which creates context with the given message.
        A custom context class can be passed.

        Parameters
        ------------
        message: :class:`.Message`
            The message to create context from.
        cls: Optional[Type]
            The optional custom class to create Context.

        Returns
        ---------
        :class:`.Context`
            The context created.
        """
        prefix = await self.get_prefix(message)

        if not cls:
            cls = Context

        ctx = cls(message=message, channel=message.channel, user=message.author, prefix=prefix)
        return ctx

    async def _dispatch(self, event: str, *args, **kwargs):
        await self._ws._dispatch(event, *args, **kwargs)

    async def _handle_checks(self, ctx, no_global_checks=False):
        command = ctx.command

        if no_global_checks:
            checks = [predicate for predicate in command._checks]
        else:
            checks = [predicate for predicate in itertools.chain(self._checks, command._checks)]

        if not checks:
            return True

        for predicate in checks:
            if inspect.iscoroutinefunction(predicate):
                result = await predicate(ctx)
            else:
                result = predicate(ctx)
            if not result:
                return predicate

            return result

    async def handle_commands(self, message, ctx=None):
        if ctx is None:
            try:
                ctx = await self.get_context(message)
            except Exception as e:
                return await self.event_error(e, message.raw_data)

        if not ctx.prefix:
            return

        content = message.content
        content = content[len(ctx.prefix)::].lstrip(' ')
        parsed = StringParser().process_string(content)

        message.clean_content = ' '.join(parsed.values())

        try:
            command = parsed.pop(0)
        except KeyError:
            return

        try:
            command = self._aliases[command]
        except KeyError:
            pass

        try:
            if command in self.commands:
                command = self.commands[command]
            elif command:
                raise CommandNotFound(f'<{command}> was not found.')
            else:
                return
        except Exception as e:
            ctx.command = None
            return await self.event_command_error(ctx, e)

        ctx.command = command
        instance = ctx.command.instance

        try:
            result = await self._handle_checks(ctx, command.no_global_checks)
        except Exception as e:
            return await self.event_command_error(ctx, e)
        else:
            if callable(result):
                return await self.event_command_error(ctx, CheckFailure(f'The command <{command.name}> failed to invoke'
                                                                        f' due to checks:: {result.__name__}'))
            elif not result:
                raise CheckFailure(f'The command <{command.name}> failed to invoke due to checks.')

        try:
            ctx.args, ctx.kwargs = await command.parse_args(instance, parsed)

            await self.global_before_hook(ctx)

            if ctx.command._before_invoke:
                await ctx.command._before_invoke(instance, ctx)

            if instance:
                await ctx.command._callback(instance, ctx, *ctx.args, **ctx.kwargs)
            else:
                await ctx.command._callback(ctx, *ctx.args, **ctx.kwargs)
        except Exception as e:
            if ctx.command.on_error:
                await ctx.command.on_error(instance, ctx, e)

            await self.event_command_error(ctx, e)

        try:
            # Invoke our after command hooks...
            if command._after_invoke:
                await ctx.command._after_invoke(ctx)
            await self.global_after_hook(ctx)
        except Exception as e:
            await self.event_command_error(ctx, e)

    async def global_before_hook(self, ctx):
        """|coro|

        Method which is called before any command is about to be invoked.

        This method is useful for setting up things before command invocation. E.g Database connections or
        retrieving tokens for use in the command.

        Parameters
        ------------
        ctx:
            The context used for command invocation.

        Examples
        ----------
        .. code:: py

            async def global_before_hook(self, ctx):
                # Make a database query for example to retrieve a specific token.
                token = db_query()

                ctx.token = token

            async def my_command(self, ctx):
                data = await self.create_clip(ctx.token, ...)

        Note
        ------
            The global_before_hook is called before any other command specific hooks.
        """
        pass

    async def global_after_hook(self, ctx):
        """|coro|

        Method which is called after any command is invoked regardless if it failed or not.

        This method is useful for cleaning up things after command invocation. E.g Database connections.

        Parameters
        ------------
        ctx:
            The context used for command invocation.

        Note
        ------
            The global_after_hook is called after the command successfully invokes.
        """
        pass

    async def event_webhook(self, data):
        """|coro|

        Event which is fired when a message from a Webhook subscription is received.

        Parameters
        ------------
        data: dict
            The webhook data as JSON.

        Warning
        ---------
            This event is only applicable when using the built in webhook server.
        """
        pass

    async def event_raw_pubsub(self, data):
        """|coro|

        Event which fires when a PubSub subscription event is received.

        Parameters
        ------------
        data:
            The raw data received from the PubSub event.

        Notes
        -------
        .. note::

            No parsing is done on the JSON and thus the data will be raw.
            A new event which parses the JSON will be released at a later date.
        """
        pass

    async def event_pubsub(self, data):
        raise NotImplementedError

    async def pubsub_subscribe(self, token: str, *topics):
        """|coro|

        Method which sends a LISTEN event over PubSub. This subscribes you to the topics provided.

        Parameters
        ------------
        token: str [Required]
            The oAuth token to use to subscribe.
        \*topics: Union[str] [Required]
            The topics to subscribe to.

        Raises
        --------
        WSConnectionFailure
            The PubSub websocket failed to connect.
        ClientError
            You reached the maximum amount of PubSub connections/Subscriptions.

        Returns
        ---------
        nonce: str
            The nonce associated with this subscription. Useful for validating responses.
        """
        nonce = uuid.uuid4().hex

        connection = await self._ws._pubsub_pool.delegate(*topics)
        await connection.subscribe(token, nonce, *topics)

        return nonce

    async def event_command_error(self, ctx, error):
        """|coro|

        Event called when an error occurs during command invocation.

        Parameters
        ------------
        ctx: :class:`.Context`
            The command context.
        error: :class:`.Exception`
            The exception raised while trying to invoke the command.
        """
        print('Ignoring exception in command: {0}:'.format(error), file=sys.stderr)
        traceback.print_exception(type(error), error, error.__traceback__, file=sys.stderr)

    async def event_mode(self, channel, user, status):
        """|coro|

        Event called when a MODE is received from Twitch.

        Parameters
        ------------
        channel: :class:`.Channel`
            Channel object relevant to the MODE event.
        user: :class:`.User`
            User object containing relevant information to the MODE.
        status: str
            The JTV status received by Twitch. Could be either o+ or o-.
            Indicates a moderation promotion/demotion to the :class:`.User`
        """
        pass

    async def event_userstate(self, user):
        """|coro|

        Event called when a USERSTATE is received from Twitch.

        Parameters
        ------------
        user: :class:`.User`
            User object containing relevant information to the USERSTATE.
        """
        pass

    async def event_raw_usernotice(self, channel, tags: dict):
        """|coro|

        Event called when a USERNOTICE is received from Twitch.
        Since USERNOTICE's can be fairly complex and vary, the following sub-events are available:

            :meth:`event_usernotice_subscription` :
            Called when a USERNOTICE Subscription or Re-subscription event is received.


        .. seealso::

            For more information on how to handle USERNOTICE's visit:
            https://dev.twitch.tv/docs/irc/tags/#usernotice-twitch-tags


        Parameters
        ------------
        channel: :class:`.Channel`	
            Channel object relevant to the USERNOTICE event.
        tags : dict
            A dictionary with the relevant information associated with the USERNOTICE.
            This could vary depending on the event.
        """
        pass

    async def event_usernotice_subscription(self, metadata):
        """|coro|

        Event called when a USERNOTICE subscription or re-subscription event is received from Twitch.

        Parameters
        ------------
        metadata: :class:`twitchio.dataclasses.NoticeSubscription`
            The object containing various metadata about the subscription event.
            For ease of use, this contains a :class:`.User` and :class:`.Channel`.
        """
        pass

    async def event_part(self, user):
        """|coro|

        Event called when a PART is received from Twitch.

        Parameters
        ------------
        user: :class:`.User`
            User object containing relevant information to the PART.
        """
        pass

    async def event_join(self, user):
        """|coro|

        Event called when a JOIN is received from Twitch.

        Parameters
        ------------
        user: :class:`.User`
            User object containing relevant information to the JOIN.
        """
        pass

    async def event_message(self, message):
        """|coro|

        Event called when a PRIVMSG is received from Twitch.

        Parameters
        ------------
        message: :class:`.Message`
            Message object containing relevant information.
        """
        await self.handle_commands(message)

    async def event_error(self, error: Exception, data=None):
        """|coro|

        Event called when an error occurs while processing data.

        Parameters
        ------------
        error: Exception
            The exception raised.
        data: str
            The raw data received from Twitch. Depending on how this is called, this could be None.

        Example
        ---------
        .. code:: py

            @bot.event
            async def event_error(error, data):
                traceback.print_exception(type(error), error, error.__traceback__, file=sys.stderr)
        """
        traceback.print_exception(type(error), error, error.__traceback__, file=sys.stderr)

    async def event_ready(self):
        """|coro|

        Event called when the Bot has logged in and is ready.

        Example
        ---------
        .. code:: py

            @bot.event
            async def event_ready():
                print(f'Logged into Twitch | {bot.nick}')
        """
        pass

    async def event_raw_data(self, data):
        """|coro|

        Event called with the raw data received by Twitch.

        Parameters
        ------------
        data: str
            The raw data received from Twitch.

        Example
        ---------
        .. code:: py

            @bot.event
            async def event_raw_data(data):
                print(data)
        """
        pass

    async def event_clearchat(self, notice):
        """|coro|

        Event called when the message in the chat be cleard.

        Example
        ---------
        .. code:: py

            @bot.event
            async def event_clearchat(notice):
                print(f'{notice.user.name}'s message has been cleard.')
        """
        pass

    def command(self, *, name: str=None, aliases: Union[list, tuple]=None, cls=Command):
        """Decorator which registers a command on the bot.

        Commands must be a coroutine.

        Parameters
        ------------
        name: str [Optional]
            The name of the command. By default if this is not supplied, the function name will be used.
        aliases: Union[list, tuple] [Optional]
            The command aliases. This must be a list or tuple.
        cls: class [Optional]
            The custom command class to override the default class. This must be similar to :class:`.Command`.
        no_global_checks : Optional[bool]
            Whether or not the command should abide by global checks. Defaults to False, which checks global checks.

        Raises
        --------
        TypeError
            Cls is not a class.
        """

        if not inspect.isclass(cls):
            raise TypeError(f'cls must be of type <class> not <{type(cls)}>')

        def decorator(func):
            cmd_name = name or func.__name__

            command = cls(name=cmd_name, func=func, aliases=aliases, instance=None)
            self.add_command(command)

            return command
        return decorator

    def event(self, func):
        """Decorator which adds an event listener to the bot.

        Example
        ---------
        .. code:: py

            @bot.event
            async def event_raw_data(data):
                print(data)

            @bot.event
            async def event_message(message):
                print(message.content)
                await bot.handle_commands(message)
        """
        if not inspect.iscoroutinefunction(func):
            raise TypeError('Events must be coroutines.')

        setattr(self, func.__name__, func)
        return func

    def check(self, func):
        """A decorator that adds a global check to the bot.

        This decorator allows regular functions or coroutines to be added to the bot.
        Global checks are ran before any other command specific checks.

        As with all other checks, the check(predicate), must contain a sole parametere of Context.

        Parameters
        ------------
        func : callable
            A regular function or coroutine to add as a global check.

        Examples
        ----------
        .. code::

            @bot.check
            async def my_global_check(self, ctx):
                return ctx.author.is_mod
        """
        self._checks.append(func)
        return func

    def add_listener(self, func, name: str=None):
        """Method which adds a coroutine as an extra listener.

        This can be used to add extra event listeners to the bot.

        Parameters
        ------------
        func: coro [Required]
            The coroutine to assign as a listener.
        name: str [Required]
            The event to register. E.g "event_message".
        """
        if not inspect.iscoroutinefunction(func):
            raise TypeError('Events must be coroutines.')

        name = name or func.__name__

        if name not in self.extra_listeners:
            self.extra_listeners[name] = [func]
        else:
            self.extra_listeners[name].append(func)

    def listen(self, event: str=None):
        """Decorator which adds a coroutine as a listener to an event.

        This can be used in place of :meth:`.event` or when more than one of the same event is required.

        Parameters
        ------------
        event: str [Optional]
            The event to listen to in the form of a string. E.g "event_message".

        Example
        ----------
        .. code:: py

            @bot.event()
            async def event_message(message):
            print(message.content)

            @bot.listen("event_message")
            async def extra_message(message):
            print(message.content)
        """
        def wrapper(func):
            self.add_listener(func, event)

            return func
        return wrapper

    async def modify_webhook_subscription(self, *, callback=None, mode, topic, lease_seconds=0, secret=None):
        """|coro|

        Creates a webhook subscription.

        Parameters
        ----------
        callback: Optional[str]
            The URL which will be called to verify the subscripton and on callback.
            If there's a webhook server running on the bot the callback will be automatically added.
        mode: :class:`.WebhookMode`
            Mode which describes whether the subscription should be created or not.
        topic: :class:`.Topic`
            Details about the subscription.
        lease_seconds: Optional[int]
            How many seconds the subscription should last. Defaults to 0, maximum is 846000.
        secret: Optional[str]
            A secret string which Twitch will use to add the `X-Hub-Signature` header to webhook requests.
            You can use this to verify the POST request came from Twitch using `sha256(secret, body)`.

        Raises
        --------
        Exception
            No callback url was specified and there is no webhook server running to retrieve a callback url from.

        HTTPException
            Bad request while modifying the subscription.
        """

        if callback is None:
            if self._webhook_server is None:
                raise Exception('No callback passed and no webhook server running to retrieve a callback url from.')

            callback = f'{self._webhook_server.external}:{self._webhook_server.port}/{self._webhook_server.callback}'

        await super().modify_webhook_subscription(
            callback=callback, mode=mode, topic=topic, lease_seconds=lease_seconds, secret=secret
        )



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2017-2021 TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import asyncio
import async_timeout
import copy
import functools
import itertools
import json
import logging
import re
import secrets
import sys
import traceback
import websockets
from typing import Union

from .backoff import ExponentialBackoff
from .dataclasses import *
from .errors import WSConnectionFailure, AuthenticationError, ClientError


log = logging.getLogger(__name__)


class PubSubPool:

    POOL_MAX = 10

    def __init__(self, loop: asyncio.BaseEventLoop, base):
        self.loop = loop
        self.base = base
        self.connections = {}

        for x in range(1, 11):
            self.connections[x] = PubSub(loop=self.loop, pool=self, node=x)

    async def delegate(self, *topics):
        for x in self.connections.values():
            if len(x._topics) == 50:
                continue
            elif len(x._topics) + len(topics) > 50:
                continue

            if x._websocket:
                return x

            await x.connect()

            if x._websocket.open:
                return x

            raise WSConnectionFailure('Unable to delegate WebSocket. Please try again.')
        raise ClientError('Maximum PubSub connections established.')


class WebsocketConnection:

    def __init__(self, bot, *, loop: asyncio.BaseEventLoop=None, **attrs):
        self._bot = bot
        self.loop = loop or asyncio.get_event_loop()

        self._token = attrs.get('irc_token')
        self._api_token = attrs.get('api_token')
        self.client_id = attrs.get('client_id')
        self._host = 'wss://irc-ws.chat.twitch.tv:443'
        self._websocket = None
        self._last_exec = None

        self._channel_cache = {}
        self._initial_channels = attrs.get('initial_channels')

        self.nick = attrs.get('nick').lower()
        self.extra_listeners = {}

        self.modes = attrs.pop('modes', ("commands", "tags", "membership"))
        self._mod_token = 0
        self._channel_token = 0
        self._rate_status = None

        self._pending_joins = {}
        self._pending_parts = {}
        self._authentication_error = False

        self.is_ready = asyncio.Event()

        self.regex = {
            "data": re.compile(
                r"^(?:@(?P<tags>\S+)\s)?:(?P<data>\S+)(?:\s)"
                r"(?P<action>[A-Z()-]+)(?:\s#)(?P<channel>\S+)"
                r"(?:\s(?::)?(?P<content>.+))?"),
            "ping": re.compile("PING (?P<content>.+)"),
            "author": re.compile(
                "(?P<author>[a-zA-Z0-9_]+)!(?P=author)"
                "@(?P=author).tmi.twitch.tv"),
            "mode": re.compile("(?P<mode>[\+\-])o (?P<user>.+)"),
            "host": re.compile(
                "(?P<channel>[a-zA-Z0-9_]+) "
                "(?P<count>[0-9\-]+)"),
            'code': re.compile(r":tmi\.twitch\.tv\s(?P<code>[0-9]{3}).*?"),
            'badges': re.compile(r"@badges=(?P<moderator>[^;]*);"
                                 r"color=(?P<color>[^;]*);"
                                 r"display-name=(?P<name>[^;]*);"
                                 r"emote-sets=(?P<emotes>[^;]*);"
                                 r"mod=(?P<mod>[^;]*);"
                                 r"subscriber=(?P<subscriber>[^;]*);"
                                 r"user-type=(?P<type>[^\s]+)\s:tmi.twitch.tv\s(?P<action>[A-Z]*)\s"
                                 r"#(?P<channel>[a-z0-9A-Z]+)"),
            "batches": re.compile(r":(?P<author>[a-zA-Z0-9_]+)!(?P=author)@(?P=author).tmi.twitch.tv"
                                  r"\s(?P<action>[A-Z()-]+)(?:\s#)(?P<channel>\S+)"),
            "nameslist": re.compile(r"(?P<author>[a-zA-Z0-9_]+).tmi.twitch.tv\s(?P<code>\S+)\s(?P=author)"
                                 r"\s=\s#(?P<channel>\S+)\s:(?P<names>.+)")}

        self._groups = ('action', 'data', 'content', 'channel', 'author')
        self._http = attrs.get('http')

        self._pubsub_pool = PubSubPool(loop=loop, base=self)

    async def _update_limit(self):

        while True:
            if self._mod_token == len(self._channel_cache):
                self._rate_status = 1
            else:
                self._rate_status = 0

            await asyncio.sleep(60)

    async def _token_update(self, status: str):
        if '+o' in status:
            self._mod_token += 1
        else:
            if self._mod_token <= 0:
                return
            self._mod_token -= 1

        if self._mod_token == len(self._channel_cache):
            self._rate_status = 1
        else:
            self._rate_status = 0

    @property
    def is_connected(self) -> bool:
        return self._websocket is not None and self._websocket.open

    async def _connect(self):
        try:
            self._websocket = await websockets.connect(self._host, timeout=30)
        except Exception as e:
            self._last_exec = e
            log.error('Websocket connection failed | %s', e)
            return

        if self.is_connected:
            # Make sure we are 100% connected
            log.debug('Sending authentication sequence payload to Twitch.')
            self.loop.create_task(self.auth_seq())

    async def wait_until_ready(self):
        await self.is_ready.wait()

    async def send_cap(self, cap: str):
        """|coro|

        Send a CAP REQ to Twitch.

        Valid caps are: commands, tags, membership

        Parameters
        ------------
        cap: str
            The cap request you wish to send to Twitch. Must be either commands, tags or membership.
        """
        log.debug('Sending CAP REQ: %s', cap)
        await self._websocket.send(f'CAP REQ :twitch.tv/{cap}')

    async def auth_seq(self, channels: Union[list, tuple]=None):
        """|coro|

        Automated Authentication process.

        Attempts to authenticate on the Twitch servers with the provided
        nickname and IRC Token (pass).

        On successful authentication, an attempt to join the provided channels is made.

        Parameters
        ------------
        channels: Union[list, tuple]
            A list or tuple of channels to attempt joining.
        """
        if not self.is_connected:
            return

        await self._websocket.send(f'PASS {self._token}\r\n')
        await self._websocket.send(f'NICK {self.nick}\r\n')

        for cap in self.modes:
            await self._websocket.send(f'CAP REQ :twitch.tv/{cap}')

        if not channels and not self._initial_channels:
            return

        channels = channels or self._initial_channels
        await self.join_channels(*channels)

    async def send_nick(self):
        """|coro|

        Sends a NICK request to the Twitch IRC Endpoint.

        This should only be used if :func:`auth_seq` was not used.
        """
        await self._websocket.send(f"NICK {self.nick}\r\n")

    async def send_privmsg(self, channel: str, content: str):
        """|coro|

        Sends a PRIVMSG to the Twitch IRC Endpoint.

        This should only be used directly in rare circumstances where a :class:`twitchio.abcs.Messageable` is not available.

        .. warning::

            This method is not directly handled by built-in rate-limits. You risk getting rate limited by twitch,
            which has a 30 minute cooldown.
        """

        content = content.replace("\n", " ")
        channel = re.sub('[#\s]', '', channel).lower()
        await self._websocket.send(f"PRIVMSG #{channel} :{content}\r\n")

        # Create a dummy message, used as a fake echo-message...
        data = f':{self.nick}!{self.nick}@{self.nick}.tmi.twitch.tv PRIVMSG(ECHO-MESSAGE) #{channel} :{content}\r\n'
        self.loop.create_task(self.process_data(data))

    async def join_channels(self, *channels: str):
        """|coro|

        Attempt to join the provided channels.

        Parameters
        ------------
        *channels : str
            An argument list of channels to attempt joining.
        """

        await asyncio.gather(*[self._join_channel(x) for x in channels])

    async def _join_channel(self, entry):
        channel = re.sub('[#\s]', '', entry).lower()
        await self._websocket.send(f'JOIN #{channel}\r\n')

        self._pending_joins[channel] = fut = self.loop.create_future()

        try:
            await asyncio.wait_for(fut, timeout=10)
        except asyncio.TimeoutError:
            self._pending_joins.pop(channel)

            raise asyncio.TimeoutError(
                f'Request to join the "{channel}" channel has timed out. Make sure the channel exists.')

    async def part_channels(self, *channels: str):
        """|coro|

        Attempt to part the provided channels.

        Parameters
        ------------
        *channels : str
            An argument list of channels to attempt parting.
        """

        await asyncio.gather(*[self._part_channel(x) for x in channels])

    async def _part_channel(self, entry):
        channel = re.sub('[#\s]', '', entry).lower()

        if channel not in self._channel_cache:
            raise ClientError(f'Request to leave the channel "{channel}" failed.')

        await self._websocket.send(f'PART #{channel}\r\n')

        self._pending_parts[channel] = fut = self.loop.create_future()

        try:
            await asyncio.wait_for(fut, timeout=10)
        except asyncio.TimeoutError:
            raise asyncio.TimeoutError(
                f'Request to leave the channel "{channel}" has timed out.')

    async def _listen(self):
        backoff = ExponentialBackoff()

        if not self.is_connected and self._last_exec:
            raise WSConnectionFailure(f'Websocket connection failure:\n\n{self._last_exec}')

        while True:
            if self._authentication_error:
                log.error('AUTHENTICATION ERROR:: Incorrect IRC Token passed.')
                raise AuthenticationError

            try:
                data = await self._websocket.recv()
            except websockets.ConnectionClosed:
                retry = backoff.delay()
                log.info('Websocket closed: Retrying connection in %s seconds...', retry)

                await asyncio.sleep(retry)
                await self._connect()
                continue

            await self._dispatch('raw_data', data)

            _task = self.loop.create_task(self.process_data(data))
            _task.add_done_callback(functools.partial(self._task_callback, data))

    def _task_callback(self, data, task):
        exc = task.exception()

        if isinstance(exc, AuthenticationError):
            self._authentication_error = True
        elif exc:
            self.loop.create_task(self.event_error(exc, data))

    async def process_ping(self, resp: str):
        await self._websocket.send(f"PONG {resp}\r\n")

    async def process_data(self, data):
        data = data.strip()

        try:
            code = int(self.regex['code'].match(data).group('code'))
        except AttributeError:
            code = None

        if code == 376 or code == 1:
            log.info('Successfully logged onto Twitch WS | %s', self.nick)

            futures = [fut for chan, fut in self._pending_joins.items() if chan.lower() in
                       [re.sub('[#\s]', '', c).lower() for c in self._initial_channels]]

            for fut in futures:
                try:
                    fut.exception()
                except asyncio.InvalidStateError:
                    pass

                if fut.done():
                    futures.remove(fut)

            if futures:
                await asyncio.wait(futures)

            await self._dispatch('ready')
            self.is_ready.set()

        elif data == ':tmi.twitch.tv NOTICE * :Login authentication failed' or\
                data == ':tmi.twitch.tv NOTICE * :Improperly formatted auth':
            log.warning('Authentication failed | %s', self._token)
            raise AuthenticationError('Websocket Authentication Failure... Check your token and nick.')

        _groupsdict = {}

        if data.startswith("PING"):
            match = self.regex["ping"]
        else:
            match = self.regex["data"]

        result = match.match(data)

        badges = self.regex['badges'].match(data)

        if badges:
            badges = {'name': badges.group('name'), 'mod': badges.group('mod'), 'action': badges.group('action'),
                      'channel': badges.group('channel')}

        try:
            tags = result.group("tags")

            tagdict = {}
            for tag in str(tags).split(";"):
                t = tag.split("=")
                if t[1].isdecimal():
                    t[1] = int(t[1])
                tagdict[t[0]] = t[1]
            tags = tagdict
        except (AttributeError, IndexError, KeyError):
            tags = None

        for group in self._groups:
            try:
                res = result.group(group)
                _groupsdict[group] = res
            except (AttributeError, KeyError, IndexError):
                pass

        await self.process_actions(data, _groupsdict, badges, tags)

    async def process_actions(self, raw: str, groups: dict, badges: dict, tags: dict=None):
        # todo add remaining actions, docs

        # Make sure the batched JOIN and PART events get sent...
        for match in self.regex['batches'].finditer(raw):
            if match.groups()[1] == 'JOIN':
                self.loop.create_task(self.join_action(match.groups()[2], match.groups()[0], tags))
            elif match.groups()[1] == 'PART':
                self.loop.create_task(self.part_action(match.groups()[2], match.groups()[0], tags))

        # Fill the channel cache with initial viewers...
        names = self.regex['nameslist'].finditer(raw)
        for match in names:
            if match.group('code') == '353':
                for name in match.group('names').split(' '):
                    self.loop.create_task(self.join_action(match.groups()[2], name, tags))

        action = groups.pop('action', None)
        data = groups.pop('data', None)
        content = groups.pop('content', None)
        channel = groups.pop('channel', None)

        if not action and badges:
            action = badges['action']
        elif not action:
            action = 'PING'

        try:
            author = self.regex["author"].match(data).group("author")
        except Exception:
            author = None

        if channel:
            try:
                channel = self._channel_cache[channel]['channel']
            except KeyError:
                channel = Channel(name=channel, ws=self, http=self._http)

        try:
            user = User(author=author, channel=channel or None, tags=tags, ws=self._websocket)
        except (TypeError, KeyError):
            user = None

        try:
            message = Message(author=user, content=content, channel=channel, raw_data=raw, tags=tags)
        except (TypeError, KeyError):
            message = None

        if action == 'RECONNECT':
            # TODO Disconnection/Reconnection Logic.
            return

        elif action == 'JOIN':
            pass

        elif action == 'PART':
            pass

        elif action == 'PING':
            log.debug('ACTION:: PING')
            await self.process_ping(content)

        elif action == 'PRIVMSG':
            await self._dispatch('message', message)
        elif action == 'PRIVMSG(ECHO-MESSAGE)':
            message.echo = True
            message._channel = copy.copy(message.channel)
            message.channel._echo = True

            await self._dispatch('raw_data', raw)
            await self._dispatch('message', message)

        elif action == 'USERNOTICE':
            await self._dispatch('raw_usernotice', channel, tags)

            if tags['msg-id'] in ('sub', 'resub'):
                user = User(author=tags['login'], channel=channel, tags=tags, ws=self._websocket)
                notice = NoticeSubscription(channel=channel, user=user, tags=tags)

                await self._dispatch('usernotice_subscription', notice)

        elif action == 'USERSTATE':
            log.debug('ACTION:: USERSTATE')

            if not user or not user.name:
                if badges:
                    user = User(author=badges['name'],
                                channel=Channel(name=badges['channel'], ws=self, http=self._http) or None,
                                tags=tags,
                                ws=self._websocket,
                                mod=badges['mod'])
                else:
                    return

            if user._name.lower() == self.nick.lower():
                try:
                    self._channel_cache[channel.name]['bot'] = user
                except KeyError:
                    self._channel_cache[channel.name] = {'channel': channel, 'bot': user}

            await self._dispatch('userstate', user)

        elif action == 'MODE':
            log.debug('ACTION:: MODE')

            mdata = re.match(r':jtv MODE #(?P<channel>.+?[a-z0-9])\s(?P<status>[\+\-]o)\s(?P<user>.*[a-z0-9])', raw)
            mstatus = mdata.group('status')

            user = User(author=mdata.group('user'), channel=channel, tags=tags, ws=self._websocket)

            if user._name.lower() == self.nick.lower():
                await self._token_update(mstatus)
                try:
                    self._channel_cache[channel.name]['bot'] = user
                except KeyError:
                    self._channel_cache[channel.name] = {'channel': channel, 'bot': user}

            await self._dispatch('mode', channel, user, mstatus)

        elif action == 'CLEARCHAT': #新增 被ban事件
            log.debug('ACTION:: CLEARCHAT')

            user = User(author=content, channel=channel, tags=tags, ws=self._websocket)
            notice = ClearChat(channel=channel, user=user, tags=tags)

            await self._dispatch('clearchat', notice)

    async def join_action(self, channel: str, author: str, tags):
        log.debug('ACTION:: JOIN: %s', channel)

        if author == self.nick:
            chan_ = Channel(name=channel, ws=self, http=self._http)
            user = User(author=author, channel=chan_, tags=tags, ws=self._websocket)

            self._channel_cache[channel] = {'channel': chan_, 'bot': user}

            if channel in self._pending_joins:
                self._pending_joins[channel].set_result(None)
                self._pending_joins.pop(channel)

            self._channel_token += 1

        try:
            cache = self._channel_cache[channel]['channel']._users
        except KeyError as e:
            raise ClientError("The \"nick\" value passed to the constructor does not match the user we are logged in as") from e

        try:
            user = cache[author.lower()]
        except KeyError:
            channel = self._channel_cache[channel]['channel']
            user = User(author=author, channel=channel, tags=tags, ws=self._websocket)

            cache[author.lower()] = user

        await self._dispatch('join', user)

    async def part_action(self, channel: str, author: str, tags):
        log.debug('ACTION:: PART: %s', channel)

        if author == self.nick:
            self._channel_cache.pop(channel)
            self._channel_token -= 1

            if self._pending_parts:
                self._pending_parts[channel].set_result(None)
                self._pending_parts.pop(channel)
        try:
            cache = self._channel_cache[channel]['channel']._users
        except KeyError:
            channel = Channel(name=channel, ws=self, http=self._http)
            user = User(author=author, channel=channel or None, tags=tags, ws=self._websocket)
        else:
            try:
                user = cache.pop(author.lower())
            except KeyError:
                channel = Channel(name=channel, ws=self, http=self._http)
                user = User(author=author, channel=channel or None, tags=tags, ws=self._websocket)

        await self._dispatch('part', user)

    async def _dispatch(self, event: str, *args, **kwargs):
        log.debug('Dispatching event: %s', event)

        func = getattr(self._bot, f'event_{event}')
        self.loop.create_task(func(*args, **kwargs))

        listeners = getattr(self._bot, 'extra_listeners', None)
        if not listeners:
            return

        extras = listeners.get(f'event_{event}', [])
        ret = await asyncio.gather(*[e(*args, **kwargs) for e in extras])

        for e in ret:
            if isinstance(e, Exception):
                self.loop.create_task(self.event_error(e))

    async def event_error(self, error: Exception, data: str=None):
        traceback.print_exception(type(error), error, error.__traceback__, file=sys.stderr)

    def teardown(self):
        if self._bot._webhook_server:
            self._bot._webhook_server.stop()

        self._websocket.close()


class PubSub:

    __slots__ = ('loop', '_pool', '_node', '_subscriptions', '_topics', '_websocket', '_timeout', '_last_result',
                 '_listener')

    def __init__(self, loop: asyncio.BaseEventLoop, pool: PubSubPool, node: int):
        self.loop = loop
        self._pool = pool
        self._node = node
        self._topics = []
        self._websocket = None
        self._timeout = asyncio.Event()

        self._last_result = None

        loop.create_task(self.handle_ping())
        self._listener = None

    @property
    def node(self) -> int:
        return self._node

    async def reconnection(self):
        backoff = ExponentialBackoff()
        self._listener.cancel()

        while True:
            retry = backoff.delay()
            log.info('PubSub Websocket closed: Retrying connection in %s seconds...', retry)

            await self.connect()

            if self._websocket is not None and self._websocket.open:
                for topic in self._topics:
                    await self.resub(topic[1], topic[0])
                return

            await asyncio.sleep(retry)

    async def connect(self):
        try:
            self._websocket = await websockets.connect('wss://pubsub-edge.twitch.tv')
        except Exception as e:
            self._last_result = e
            log.error('PubSub websocket connection failed | %s', e)

            raise WSConnectionFailure('PubSub websocket connection failed. Check logs for details.')

        log.info('PubSub %s connection successful', self.node)
        self._listener = self.loop.create_task(self.listen())

    @staticmethod
    def generate_jitter():
        # Generates a random number between around 1 and 10
        jitter = 0

        while jitter == 11 or jitter == 0:
            bites = secrets.token_bytes(2)
            number = itertools.accumulate(bites)
            jitter = int(sum(number) / 2 ** 6)

        return jitter

    async def handle_ping(self):
        while True:
            jitter = self.generate_jitter()
            await asyncio.sleep(240 + jitter)
            self._timeout.clear()

            if not self._websocket:
                continue
            if self._websocket.open:
                log.debug('PubSub %s Sending PING payload.', self.node)
                await self._websocket.send(json.dumps({"type": "PING"}))
            else:
                continue

            try:
                async with async_timeout.timeout(10):
                    await self._timeout.wait()
            except asyncio.TimeoutError:
                self.loop.create_task(self.reconnection())

    async def listen(self):
        while True:
            try:
                data = json.loads(await self._websocket.recv())
                self.loop.create_task(self._pool.base._dispatch('raw_pubsub', data))
            except websockets.ConnectionClosed:
                return self.loop.create_task(self.reconnection())

            if data['type'] == 'PONG':
                log.debug('PubSub %s received PONG payload.', self.node)
                self._timeout.set()
            elif data['type'] == 'RECONNECT':
                log.debug('PubSub %s received RECONNECT payload... Attempting reconnection', self.node)
                self.loop.create_task(self.reconnection())

            # self.loop.create_task(self._pool.base._dispatch('pubsub', data))

    async def subscribe(self, token: str, nonce: str, *topics: str):
        for t in topics:
            self._topics.append((t, token))

        payload = {"type": "LISTEN",
                   "nonce": nonce,
                   "data": {"topics": [*topics],
                            "auth_token": token}}

        await self._websocket.send(json.dumps(payload))

    async def resub(self, token: str, topic: str):
        payload = {"type": "LISTEN",
                   "data": {"topics": [topic],
                            "auth_token": token}}

        await self._websocket.send(json.dumps(payload))




# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2017-2021 TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

__all__ = ('Message', 'Channel', 'User', 'Context', 'NoticeSubscription', 'ClearChat')


import datetime
import time
from typing import *

from .abcs import Messageable
from .errors import EchoMessageWarning, HTTPException, Unauthorized


class Message:

    __slots__ = ('_author', '_channel', '_raw_data', 'content', 'clean_content', '_tags', '_timestamp', 'echo')

    def __init__(self, **attrs):
        self._author = attrs.pop('author', None)
        self._channel = attrs.pop('channel', None)
        self._raw_data = attrs.pop('raw_data', None)
        self.content = attrs.pop('content', None)
        self.clean_content = attrs.pop('clean_content', None)
        self._tags = attrs.pop('tags', None)
        self.echo = False

        try:
            self._timestamp = self._tags['tmi-sent-ts']
        except (TypeError, KeyError):
            self._timestamp = time.time()

    @property
    def author(self) -> 'User':  # stub
        """The User object associated with the Message."""
        return self._author

    @property
    def channel(self) -> 'Channel':  # stub
        """The Channel object associated with the Message."""
        return self._channel

    @property
    def raw_data(self) -> str:
        """The raw data received from Twitch for this Message."""
        return self._raw_data

    @property
    def tags(self) -> Optional[dict]:
        """The tags associated with the Message.

        Could be None.
        """
        return self._tags

    @property
    def timestamp(self) -> datetime.datetime.timestamp:
        """The Twitch timestamp for this Message.

        Returns
        ---------
        timestamp:
            UTC datetime object of the Twitch timestamp.
        """
        timestamp = datetime.datetime.utcfromtimestamp(int(self._timestamp) / 1000)
        return timestamp


class Channel(Messageable):

    __slots__ = ('_channel', '_ws', '_http', '_echo', '_users')

    def __init__(self, name, ws, http):
        self._channel = name
        self._http = http
        self._ws = ws
        self._echo = False
        self._users = {}

    def __str__(self):
        return self._channel

    @property
    def name(self) -> str:
        """The channel name."""
        return self._channel

    @property
    def chatters(self) -> list:
        """The channel's chatters."""
        return list(self._users.values())

    def _get_channel(self) -> Tuple[str, None]:
        return self.name, None

    def _get_method(self) -> str:
        return self.__class__.__name__

    @property
    def _get_socket(self):  # stub
        if self._echo is True:
            raise EchoMessageWarning('Unable to respond to Echo-Messages.')

        return self._ws

    async def get_stream(self) -> dict:
        """|coro|

        Method which retrieves stream information on the channel, provided it is active (Live).

        Returns
        ---------
        dict:
            Dict containing active streamer data. Could be None if the stream is not live.

        Raises
        --------
        HTTPException
            Bad request while fetching streams.
        """

        data = await self._http.get_streams(channels=[self.name])

        try:
            return data[0]
        except IndexError:
            pass

    async def get_custom_rewards(self, token: str, broadcaster_id: int, *, only_manageable=False, ids: List[int]=None) -> List["CustomReward"]:
        """
        Fetches the channels custom rewards (aka channel points) from the api.
        Parameters
        ----------
        token : :class:`str`
            The users oauth token.
        broadcaster_id : :class:`int`
            The id of the broadcaster.
        only_manageable : :class:`bool`
            Whether to fetch all rewards or only ones you can manage. Defaults to false.
        ids : List[:class:`int`]
            An optional list of reward ids

        Returns
        -------

        """
        try:
            data = await self._http.get_rewards(token, broadcaster_id, only_manageable, ids)
        except Unauthorized as error:
            raise Unauthorized("The given token is invalid", "", 401) from error
        except HTTPException as error:
            status = error.args[2]
            if status == 403:
                raise HTTPException("The custom reward was created by a different application, or channel points are "
                                    "not available for the broadcaster (403)", error.args[1], 403) from error
            raise
        else:
            return [CustomReward(self._http, x, self) for x in data['data']]


class User:

    __slots__ = ('_name', '_channel', '_tags', 'display_name', '_id', 'type',
                 '_colour', 'subscriber', 'turbo', '_badges', '_ws', '_mod')

    def __init__(self, ws, **attrs):
        self._name = attrs.pop('author', None)
        self._channel = attrs.pop('channel', self._name)
        self._tags = attrs.pop('tags', None)
        self._ws = ws

        if not self._tags:
            self._tags = {}

        self.display_name = self._tags.get('display-name', self._name)
        self._id = int(self._tags.get('user-id', 0))
        self.type = self._tags.get('user-type', 'Empty')
        self._colour = self._tags.get('color', None)
        self.subscriber = self._tags.get('subscriber', None)
        self.turbo = self._tags.get('turbo', None)

        self._badges = {}
        badges = self._tags.get('badges', None)
        if badges:
            for chunk in badges.split(','):
                k, _, v = chunk.partition('/')
                self._badges[k] = v

        self._mod = int(self._tags.get('mod', 0)) if self._tags else attrs.get('mod', 0)

    def __repr__(self):
        return '<User name={0.name} channel={0._channel}>'.format(self)

    def __eq__(self, other):
        return other == self.name

    def __hash__(self):
        return hash(self.name)

    @property
    def name(self) -> str:
        """The user's name."""
        return self._name

    @property
    def id(self) -> int:
        """The user's ID.

         Could be 0 if no Tags were received."""
        return self._id

    @property
    def channel(self) -> Channel:
        """The channel object associated with the User.

        .. note::

            The channel will be valid for the data which triggered the Event. It is possible the
            user could be in multiple channels. E.g: The User BobRoss sends a message from the Channel ArtIsCool.
            The Channel object received will be ArtIsCool.
        """
        return self._channel

    @property
    def colour(self) -> Optional[str]:
        """The user's colour.

        Could be None if no Tags were received.
        """
        return self._colour

    @property
    def color(self) -> Optional[str]:
        """An American-English alias to colour."""
        return self.colour

    @property
    def is_turbo(self) -> bool:
        """A boolean indicating whether the User is Turbo.

        Could be None if no Tags were received.
        """
        return self.turbo

    @property
    def is_subscriber(self) -> bool:
        """A boolean indicating whether the User is a subscriber of the current channel.

        Could be None if no Tags were received.
        """
        return self.subscriber

    @property
    def badges(self) -> dict:
        """The badges associated with the User.

        Could be an empty Dict if no Tags were received.
        """
        return self._badges

    @property
    def tags(self) -> dict:
        """The Tags received for the User.

        Could be an empty Dict if no tags were received.
        """
        return self._tags

    @property
    def prediction(self) -> Optional[str]:
        """
        the chatters current prediction.

        Returns
        --------
        Optional[:class:`str`] Either blue, pink, or None
        """
        if "blue-1" in self._badges:
            return "blue"
        elif "pink-2" in self._badges:
            return "pink"

        return None

    @property
    def is_mod(self) -> bool:
        """A boolean indicating whether the User is a moderator of the current channel."""
        if self._mod == 1:
            return True
        if self.channel.name == self.name.lower():
            return True
        else:
            return False


class Context(Messageable):
    """
    The context of which a command is invoked under.

    Attributes
    ------------
    author : :class:`.User`
        The author of the command.
    prefix : str
        The prefix associated with the command.
    message : :class:`.Message`
        The message associated with the command.
    channel : :class:`.Channel`
        The channel associated with the command.
    command : :class:`twitchio.ext.core.Command`
        The command which was invoked.


    .. note::
        Context is only available through the commands extension.
    """

    def __init__(self, message: Message, channel: Channel, user: User, **attrs):
        self.message = message
        self.channel = channel
        self.content = message.content
        self.author = user
        self.prefix = attrs.get('prefix', None)

        self._echo = self.channel._echo
        self._ws = self.channel._ws

        self.command = attrs.get('Command', None)
        self.args = attrs.get('args', None)
        self.kwargs = attrs.get('kwargs', None)

    def _get_channel(self) -> Tuple[str, None]:
        return self.channel.name, None

    def _get_method(self) -> str:
        return self.__class__.__name__

    @property
    def _get_socket(self):  # stub
        if self._echo is True:
            raise EchoMessageWarning('Unable to respond to Echo-Messages.')

        return self._ws

    async def get_stream(self) -> dict:
        """|coro|

        Method which retrieves stream information on the channel stored in Context, provided it is active (Live
        Returns
        ---------
        dict
            Dict containing active streamer data. Could be None if the stream is not live.

        Raises
        --------
        HTTPException
            Bad request while fetching streams.
        """
        return await self.channel.get_stream()


class NoticeSubscription:
    """
    The Dataclass sent to `event_usernotice_subscription` events.

    Attributes
    ------------
    channel : :class:`.Channel`
        The channel associated with the subscription event.
    user : :class:`.User`
        The user associated with the subscription event.
    tags : dict
        The raw tags dict associated with the subscription event.
    cumulative_months : Optional[int]
        The total number of months the user has subscribed. Could be None if not provided.
    share_streak : bool
        Boolean indicating whether users want their streaks to be shared.
    streak_months : Optional[int]
        The number of consecutive months the user has subscribed. Could be None if not provided.
        This is 0 if ``share_streak`` is False.
    sub_plan : str
        The type of subscription plan being used.
        Valid values: Prime, 1000, 2000, 3000.

        1000, 2000, and 3000 refer to the first, second, and third levels of paid subscriptions,
        respectively (currently $4.99, $9.99, and $24.99).
    sub_plan_name : str
        The display name of the subscription plan.
        This may be a default name or one created by the channel owner.
    """

    def __init__(self, *, channel: Channel, user: User, tags: dict):
        self.channel = channel
        self.user = user

        self.tags = tags

        self.cumulative_months = tags.get('msg-param-cumulative-months', None)
        if self.cumulative_months:
            self.cumulative_months = int(self.cumulative_months)

        self.share_streak = bool(tags['msg-param-should-share-streak'])

        self.streak_months = tags.get('msg-param-streak-months', None)
        if self.streak_months:
            self.streak_months = int(self.streak_months)

        self.sub_plan = tags['msg-param-sub-plan']
        self.sub_plan_name = tags['msg-param-sub-plan-name']

class CustomReward:
    """
    Represents a Custom Reward object, as given by the api. Use :ref:`User.get_custom_rewards` to fetch these
    """
    __slots__ = "_http", "_channel", "id", "image", "background_color", "enabled", "cost", "title", "prompt", \
                "input_required", "max_per_stream", "max_per_user_stream", "cooldown", "paused", "in_stock", \
                "redemptions_skip_queue", "redemptions_current_stream", "cooldown_until", "_broadcaster_id"

    def __init__(self, http, obj: dict, channel: Channel):
        self._http = http
        self._channel = channel
        self._broadcaster_id = obj['broadcaster_id']
        
        self.id = obj['id']
        self.image = obj['image']['url_1x'] if obj['image'] else obj['default_image']['url_1x']
        self.background_color = obj['background_color']
        self.enabled = obj['is_enabled']
        self.cost = obj['cost']
        self.title = obj['title']
        self.prompt = obj['prompt']
        self.input_required = obj['is_user_input_required']
        self.max_per_stream = obj['max_per_stream_setting']['is_enabled'], obj['max_per_stream_setting']['max_per_stream']
        self.max_per_user_stream = obj['max_per_user_per_stream_setting']['is_enabled'], \
                                    obj['max_per_user_per_stream_setting']['max_per_user_per_stream']
        self.cooldown = obj['global_cooldown_setting']['is_enabled'], obj['global_cooldown_setting']['global_cooldown_seconds']
        self.paused = obj['paused']
        self.in_stock = obj['is_in_stock']
        self.redemptions_skip_queue = obj['should_redemptions_skip_request_queue']
        self.redemptions_current_stream = obj['redemptions_redeemed_current_stream']
        self.cooldown_until = obj['cooldown_expires_at']


    async def edit(
            self,
            token: str,
            title: str = None,
            prompt: str = None,
            cost: int = None,
            background_color: str = None,
            enabled: bool = None,
            input_required: bool = None,
            max_per_stream_enabled: bool = None,
            max_per_stream: int = None,
            max_per_user_per_stream_enabled: bool = None,
            max_per_user_per_stream: int = None,
            global_cooldown_enabled: bool = None,
            global_cooldown: int = None,
            paused: bool = None,
            redemptions_skip_queue: bool = None
    ):
        """
        Edits the reward. Note that apps can only modify rewards they have made.
        
        Parameters
        -----------
        token: the bearer token for the channel of the reward
        title: the new title of the reward
        prompt: the new prompt for the reward
        cost: the new cost for the reward
        background_color: the new background color for the reward
        enabled: whether the reward is enabled or not
        input_required: whether user input is required or not
        max_per_stream_enabled: whether the stream limit should be enabled
        max_per_stream: how many times this can be redeemed per stream
        max_per_user_per_stream_enabled: whether the user stream limit should be enabled
        max_per_user_per_stream: how many times a user can redeem this reward per stream
        global_cooldown_enabled: whether the global cooldown should be enabled
        global_cooldown: how many seconds the global cooldown should be
        paused: whether redemptions on this reward should be paused or not
        redemptions_skip_queue: whether redemptions skip the request queue or not

        Returns
        --------
        :class:`CustomReward` itself.
        """

        try:
            data = await self._http.update_reward(
                token,
                self._broadcaster_id,
                self._id,
                title,
                prompt,
                cost,
                background_color,
                enabled,
                input_required,
                max_per_stream_enabled,
                max_per_stream,
                max_per_user_per_stream_enabled,
                max_per_user_per_stream,
                global_cooldown_enabled,
                global_cooldown,
                paused,
                redemptions_skip_queue
            )
        except Unauthorized as error:
            raise Unauthorized("The given token is invalid", "", 401) from error
        except HTTPException as error:
            status = error.args[2]
            if status == 403:
                raise HTTPException("The custom reward was created by a different application, or channel points are "
                                    "not available for the broadcaster (403)", error.args[1], 403) from error
            raise
        else:
            for reward in data['data']:
                if reward['id'] == self._id:
                    self.__init__(self._http, reward, self._channel)
                    break

        return self

    async def delete(self, token: str):
        """
        Deletes the custom reward

        Parameters
        ----------
        token: :class:`str` the oauth token of the target channel

        Returns
        --------
        None
        """
        try:
            await self._http.delete_reward(token, self._broadcaster_id, self._id)
        except Unauthorized as error:
            raise Unauthorized("The given token is invalid", "", 401) from error
        except HTTPException as error:
            status = error.args[2]
            if status == 403:
                raise HTTPException("The custom reward was created by a different application, or channel points are "
                                    "not available for the broadcaster (403)", error.args[1], 403) from error
            raise

    async def get_redemptions(self, token: str, status: str, sort: str = None):
        """
        Gets redemptions for this reward

        Parameters
        -----------
        token: :class:`str` the oauth token of the target channel
        status: :class:`str` one of UNFULFILLED, FULFILLED or CANCELED
        sort: :class:`str` the order redemptions are returned in. One of OLDEST, NEWEST. Default: OLDEST.
        """
        try:
            data = await self._http.get_reward_redemptions(token, self._broadcaster_id, self._id, status=status, sort=sort)
        except Unauthorized as error:
            raise Unauthorized("The given token is invalid", "", 401) from error
        except HTTPException as error:
            status = error.args[2]
            if status == 403:
                raise HTTPException("The custom reward was created by a different application, or channel points are "
                                    "not available for the broadcaster (403)", error.args[1], 403) from error
            raise
        else:
            return [CustomRewardRedemption(x, self._http, self) for x in data['data']]


class CustomRewardRedemption:
    __slots__ = "_http", "_broadcaster_id", "id", "user_id", "user_name", "input", "status", "redeemed_at", "reward"
    def __init__(self, obj, http, parent):
        self._http = http
        self._broadcaster_id = obj['broadcaster_id']
        self.id = obj['id']
        self.user_id = int(obj['user_id'])
        self.user_name = obj['user_name']
        self.input = obj['user_input']
        self.status = obj['status']
        self.redeemed_at = datetime.datetime.fromisoformat(obj['redeemed_at'])
        self.reward = parent or obj['reward']

    async def fulfill(self, token: str):
        """
        marks the redemption as fulfilled

        Parameters
        ----------
        token: :class:`str` the token of the target channel

        Returns
        --------
        itself.
        """
        reward_id = self.reward.id if isinstance(self.reward, CustomReward) else self.reward['id']
        try:
            data = await self._http.update_reward_redemption_status(token, self._broadcaster_id, self.id, reward_id, "FULFILLED")
        except Unauthorized as error:
            raise Unauthorized("The given token is invalid", "", 401) from error
        except HTTPException as error:
            status = error.args[2]
            if status == 403:
                raise HTTPException("The custom reward was created by a different application, or channel points are "
                                    "not available for the broadcaster (403)", error.args[1], 403) from error
            raise
        else:
            self.__init__(data['data'], self._http, self.reward if isinstance(self.reward, CustomReward) else None)
            return self

    async def refund(self, token: str):
        """
        marks the redemption as cancelled

        Parameters
        ----------
        token: :class:`str` the token of the target channel

        Returns
        --------
        itself.
        """
        reward_id = self.reward.id if isinstance(self.reward, CustomReward) else self.reward['id']
        try:
            data = await self._http.update_reward_redemption_status(token, self._broadcaster_id, self.id, reward_id,
                                                                    False)
        except Unauthorized as error:
            raise Unauthorized("The given token is invalid", "", 401) from error
        except HTTPException as error:
            status = error.args[2]
            if status == 403:
                raise HTTPException("The custom reward was created by a different application, or channel points are "
                                    "not available for the broadcaster (403)", error.args[1], 403) from error
            raise
        else:
            self.__init__(data['data'], self._http, self.reward if isinstance(self.reward, CustomReward) else None)
            return self

class ClearChat: # Add class for event_ban
    """
    The Dataclass sent to `event_ban` events.
    Attributes
    ------------
    channel : :class:`.Channel`
        The channel associated with the subscription event.
    user : :class:`.User`
        The user associated with the subscription event.
    tags : dict
        The raw tags dict associated with the subscription event.
    ban_duration : Optional[int]
        Duration of the timeout, in seconds.Could be None if the ban is permanent.
    """

    def __init__(self, *, channel: Channel, user: User, tags: dict):
        self.channel = channel
        self.user = user

        self.tags = tags

        self.ban_duration = tags.get('ban-duration', None)
        if self.ban_duration:
            self.ban_duration = int(self.ban_duration)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2017-2021 TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import asyncio
import time


class RateBucket:

    HTTPLIMIT = 30
    IRCLIMIT = 20
    MODLIMIT = 100

    HTTP = 60
    IRC = 30

    def __init__(self, *, method: str):
        self.method = method

        if method == 'irc':
            self.reset_time = self.IRC
            self.limit = self.IRCLIMIT
        elif method == 'mod':
            self.reset_time = self.IRC
            self.limit = self.MODLIMIT
        else:
            self.reset_time = self.HTTP
            self.limit = self.HTTPLIMIT

        self.tokens = 0
        self._reset = time.time() + self.reset_time

    @property
    def limited(self):
        return self.tokens == self.limit

    def reset(self):
        self.tokens = 0
        self._reset = time.time() + self.reset_time

    def update(self, *, reset=None, remaining=None):
        now = time.time()

        if self._reset <= now:
            self.reset()

        if reset:
            self._reset = int(reset)

        if remaining:
            self.tokens = self.limit - int(remaining)
        else:
            self.tokens += 1

    async def wait_reset(self):
        now = time.time()

        await asyncio.sleep(self._reset - now)
        self.reset()
